%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  $RCSfile: tokenization.pl,v $
%% $Revision: 1.9 $
%%     $Date: 2004/01/02 02:40:13 $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Defines tokenization to be used for this grammar:
%%
%% Author: Detmar Meurers 
%%         with decapitalization rules for Polish by Adam Przepiorkowski
%%         and idea to use sentence final punctuation by Stefan Mueller
%%
%% Basic idea: Translate a string into words, taking any number 
%%    of spaces as delimiting words, e.g.
%%    "ab"     -> [ab]
%%    "ab cd"  -> [ab,cd]
%%    "ab  cd" -> [ab,cd]
%%
%%    Note that this is equivalent to
%%    [97,98]              -> [ab]
%%    [97,98,32,99,100]    -> [ab,cd]
%%    [97,98,32,32,99,100] -> [ab,cd]
%%
%% Specific cases:
%% - decapitalizes all capital letters
%% - removes the sentence internal punctuation symbols: ; ! , . ?  
%% - a ' ends a word and starts the next word (even if the last symbol)
%%   e.g. "John's" -> [john,'\'s']
%%        "Jons'" -> [Johns,'\'']
%% - treats sentence final . ! and ? as referring to the start symbols
%%   defined by decl_symbol/1, imp_symbol/1 and que_symbol/1;
%%   if none of these ends the sentence, root_symbol/1 is used
%%

% Until we have a real treatment of lexical/syntactic ambiguity, the following
% stipulates which multi word items are treated lexically.

% The following negations are treated lexically instead of syntactically:
multi_word_entry([isn,'''t'|X],X,[isnt|Y],Y).
multi_word_entry([aren,'''t'|X],X,[arent|Y],Y).
multi_word_entry([wasn,'''t'|X],X,[wasnt|Y],Y).
multi_word_entry([weren,'''t'|X],X,[werent|Y],Y).
multi_word_entry([can,'''t'|X],X,[cant|Y],Y).
multi_word_entry([couldn,'''t'|X],X,[couldnt|Y],Y).
multi_word_entry([shouldn,'''t'|X],X,[shouldnt|Y],Y).
multi_word_entry([won,'''t'|X],X,[wont|Y],Y).
multi_word_entry([wouldn,'''t'|X],X,[wouldnt|Y],Y).
multi_word_entry([don,'''t'|X],X,[dont|Y],Y).
multi_word_entry([doesn,'''t'|X],X,[doesnt|Y],Y).
multi_word_entry([didn,'''t'|X],X,[didnt|Y],Y).
multi_word_entry([hasn,'''t'|X],X,[hasnt|Y],Y).
multi_word_entry([haven,'''t'|X],X,[havent|Y],Y).
multi_word_entry([hadn,'''t'|X],X,[hadnt|Y],Y).

multi_word_entry([had,better|X],X,[had_better|Y],Y).
multi_word_entry([palo,alto|X],X,[palo_alto|Y],Y).
multi_word_entry([in,order,to|X],X,[in_order_to|Y],Y).
multi_word_entry([each,other|X],X,[each_other|Y],Y).
multi_word_entry([as,soon,as|X],X,[as_soon_as|Y],Y).
multi_word_entry([a,dozen|X],X,[a_dozen|Y],Y).

% Numbers and dates should get a special treatment, which wasn't part of
% the ERG grammar MERGE is based on. So here's just a simple mapping of some examples:
multi_word_entry(['15'|X],X,[fifteen|Y],Y).
multi_word_entry(['1986'|X],X,[nineteen_eightysix|Y],Y).
multi_word_entry(['1984'|X],X,[nineteen_eightyfour|Y],Y).
multi_word_entry(['1st'|X],X,[first|Y],Y).
multi_word_entry(['1/3/84'|X],X,[third,of,january,nineteen_eightyfour|Y],Y).
multi_word_entry(['1-3-84'|X],X,[third,of,january,nineteen_eightyfour|Y],Y).

% Until there is a real treatment of punctuation, we ignore standalone hyphens:
multi_word_entry(['-'|X],X,Y,Y).
multi_word_entry(['--'|X],X,Y,Y).

% not yet caught here: "do not" in imperatives only is also treated as a token

tokenize_sentence_string(String,Tokenlist,Desc):-
   tokenize_sentence_string_act(String,Tokenlist0,Desc),
   catch_multiword_entries(Tokenlist0,Tokenlist).

catch_multiword_entries([],[]).

catch_multiword_entries(In,Out) :-
  multi_word_entry(In,InRest,Out,OutRest),
  !,% for now, we cut here, so no ambiguity between multiword expression vs. parse with separate words
  catch_multiword_entries(InRest,OutRest).

catch_multiword_entries([First|Rest],[First|NewRest]) :-
   catch_multiword_entries(Rest,NewRest).

% tokenize_sentence_string_act(+String,-Tokenlist,-StartSymbolDesc)
tokenize_sentence_string_act([],[],Desc) :- 
   !,
   (  current_predicate(root_symbol/1) 
   -> root_symbol(Desc)
   ;  Desc = bot
   ).
tokenize_sentence_string_act([LastChar],[],Desc) :- 
   LastChar==fullstop,
   !,
   (  current_predicate(decl_symbol/1) 
   -> decl_symbol(Desc)
   ;  Desc = bot
   )
   ;
   LastChar==exclam,
   !,
   (  current_predicate(imp_symbol/1) 
   -> imp_symbol(Desc)
   ;  Desc = bot
   )
   ;
   LastChar==question,
   !,
   (  current_predicate(que_symbol/1) 
   -> que_symbol(Desc)
   ;  Desc = bot
   ).
tokenize_sentence_string_act(String,AtomList,StartSymb) :-
   tokenize_word_string(String,AtomList0,AtomList,RestString),
   tokenize_sentence_string_act(RestString,AtomList0,StartSymb).

tokenize_word_string(String,AtomList0,AtomList,RestString) :-
   tokenize_one_word(String,WordList,RestString),
   (  WordList == []
   -> AtomList = AtomList0
   ;  atom_codes(Atom,WordList),
      AtomList = [Atom|AtomList0]
   ).

tokenize_one_word([],[],[]).                  % end of string reached
tokenize_one_word([H|T],Word,Rest) :-
   (  H == 32                        % a) space ends a word
   -> Word=[],eliminate_spaces(T,Rest)  %    and remove additional spaces
   ;  (  H == 39                     % b) ' ends a word
      -> Word =[],Rest=[quote|T]     %    and leave quote at start of next word
      ;  (  sentence_final_punct(H,NewH) 
         -> Word=[],Rest=[NewH|T]
         ;  tokenize_one_char(H,Word0,Word), % else c), map one char
            tokenize_one_word(T,Word0,Rest)      %    and continue on word
         )
      )
   ).

sentence_final_punct(H,NewH) :-
   NewH=fullstop,
   atom_codes('.',[H]), !
   ;
   NewH=exclamation,
   atom_codes('!',[H]), !
   ;
   NewH=question,
   atom_codes('?',[H]).

tokenize_one_char(quote,L,[39|L]) :- % quote is output as '
   !.
% remove punctuation except for sentence ending .,?, and ! (treated above)
tokenize_one_char(Char,L,L) :-       
   (sentence_final_punct(_,Char)
   ;atom_codes('(',[Char])
   ;atom_codes(')',[Char])
   ;atom_codes(';',[Char])
   ;atom_codes(',',[Char])),
   !.
tokenize_one_char(Upper,L,[Lower|L]) :- % letters are decapitalized
   decapitalize_char(Upper,Lower),
   !.
tokenize_one_char(Char,L,[Char|L]).

eliminate_spaces([],[]).
eliminate_spaces([H|T],R) :-
   (  H == 32 
   -> eliminate_spaces(T,R)
   ;  [H|T]=R
   ).

% ------------------------------------------------------------------------
% Converting to lower caps: This is not the fastest code, since we could
% precompute the atom_codes calls, but it's more transparent this way

decapitalize_char(X,XDecap):-   % fails for non-cap characters
   atom_codes('A',[A]),
   atom_codes('Z',[Z]),
   atom_codes( a ,[Lower_a]),
   Diff is Lower_a - A,
   ( (A =< X, X =< Z)
   -> XDecap is X + Diff        % basic capital to lower letter conversion
   ; decap_special(X,XDecap)    % specials for other character sets
   ).

% Decapitalizing German Umlauts:
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!,             
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!,                     
   atom_codes(,[Y]).
% Polish letters added by Adam P.
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).
decap_special(X,Y) :-           %  -> 
   atom_codes('',[X]),!, 
   atom_codes(,[Y]).

